import 'dart:convert';
import 'dart:io' show File;

import 'package:_fe_analyzer_shared/src/parser/parser.dart';
import 'package:_fe_analyzer_shared/src/scanner/io.dart' show readBytesFromFileSync;
import 'package:_fe_analyzer_shared/src/scanner/recover.dart';
import 'package:_fe_analyzer_shared/src/scanner/scanner.dart';
import 'package:_fe_analyzer_shared/src/scanner/token.dart';
import 'package:analyzer/source/line_info.dart';
import 'package:logging/logging.dart';
import 'package:report/parser/logger.dart';
import 'package:report/parser/stack.dart';

import 'bloc_path.dart';
import 'property_path.dart';
import 'report.dart';
import 'write_token.dart';

bool isLogger = false;
bool isTest = false;
void main() async {
  // await runAnalyze(
  //   rootPath: '/Users/dansun/Desktop/edu/ai_teacher_middle_school/src',
  //   reportPath: '/Users/dansun/Documents/github/report/tool/analyze/report.json',
  // );
  await runAnalyze(
    rootPath: '/Users/dansun/github/report',
    reportPath: '/Users/dansun/github/report/tool/analyze/report.json',
  );
}

void runAnalyze({String rootPath, reportPath}) async {
  Logger.root.level = Level.ALL; // defaults to Level.INFO
  Logger.root.onRecord.listen((record) {
    print('${record.level.name}: ${record.time}: ${record.message}');
  });
  logInfo('开始上报扫描');

  DateTime begin = DateTime.now();
  Map<String, dynamic> reportMap = await reportJson(reportPath);
  if (reportMap == null) {
    return null;
  }
  ReportData reportData = ReportData(reportMap);
  List<String> includePaths = await packagePath(rootPath);

  print('includePaths ${includePaths.toString()}');

  mainEntryPoint(await includePaths, reportData);
  logInfo('完成用时 ${DateTime.now().difference(begin)} 扫描文件 ${includePaths.length}');
}

Future<void> mainEntryPoint(List<String> arguments, ReportData reportData) async {
  for (String argument in arguments) {
    outLine(Uri.base.resolve(argument), reportData);
  }
}

void outLine(Uri uri, ReportData reportData) {
  ParserListener parserListener = ParserListener();
  parserListener.list = [];
  Parser parser = Parser(parserListener);
  ScannerResult scannerResult = scan(
    readBytesFromFileSync(uri),
    includeComments: true,
  );
  LineInfo lineInfo = LineInfo(scannerResult.lineStarts);

  parser.parseUnit(scannerResult.tokens);

  if (parserListener.list.isNotEmpty) {
    bool isUpdate = false;

    parserListener.list.forEach((element) {
      bool update = updateToken(element, reportData, parser, uri);
      if (!isUpdate && update) {
        isUpdate = update;
      }
    });
    if (isUpdate) {
      writeToken(scannerResult.tokens, uri.path, lineInfo);
    }
  }

  // printToken(scannerResult.tokens);
}

bool updateToken(
  Constructor element,
  ReportData reportData,
  Parser parser,
  Uri uri,
) {
  bool isUpdate = false;
  element.properties.forEach((Property property) {
    // Property property = element.property;
    Map<String, String> map = reportData.reportData(property?.uuidPosition?.endToken?.lexeme);
    if (map != null && map.containsKey('name')) {
      isUpdate = true;

      ///替换的部分
      if (property.namePosition != null) {
        Token token = synthesizeToken(property.namePosition.endToken.offset, '\'${map['name']}\'', TokenType.STRING);
        parser.rewriter.replaceTokenFollowing(property.namePosition.endToken.previous, token);
      } else if (property.namePosition == null && property.uuidPosition.beginToken.lexeme == 'uuid') {
        /// name : ***
        Token commaToken = SyntheticStringToken(TokenType.COMMA, ',', property.uuidPosition.endToken.offset, 0);
        commaToken = parser.rewriter.insertToken(property.uuidPosition.endToken, commaToken);

        Token nameToken = synthesizeToken(commaToken.offset, 'name', TokenType.STRING);
        nameToken = parser.rewriter.insertToken(commaToken, nameToken);

        Token colonToken = synthesizeToken(nameToken.offset, ':', TokenType.COLON);
        colonToken = parser.rewriter.insertToken(nameToken, colonToken);

        Token newToken = synthesizeToken(property.uuidPosition.endToken.offset, '\'${map['name']}\'', TokenType.STRING);
        parser.rewriter.insertToken(colonToken, newToken);
      }
      if (element.constructorBegin != null && map.containsKey('comment')) {
        if (element.constructorBegin is SimpleToken) {
          SimpleToken token = element.constructorBegin as SimpleToken;
          String string = token?.precedingComments?.lexeme ?? '';
          Token commentToken = CommentToken(TokenType.SINGLE_LINE_COMMENT, '${string} \/\/todo ${map['comment']} ', element.constructorBegin.previous.offset);
          token.precedingComments = commentToken;
        }
      }
    }
  });

  return isUpdate;
}

class ParserListener extends Listener {
  List<Constructor> list = [];
  void logEvent(String name) {
    // print('=== name $name');
  }

  void beginLiteralString(Token token) {
    // print('beginLiteralString ${token.toString()}');
  }

  void endLiteralString(int interpolationCount, Token endToken) {
    // print('endLiteralString ${endToken.toString()}');
  }

  void beginClassDeclaration(Token begin, Token abstractToken, Token name) {}

  void handleIdentifier(Token token, IdentifierContext context) {
    if (isEndWithUuidToken(token)) {
      Token previous = token.previous;
      while (!(previous is BeginToken)) {
        previous = previous.previous;
      }
      list.add(scanTokens(previous.previous));
    }

    ///uuid 结尾的部分
  }
}

void ScanToken(Token token) {
  StringBuffer a = StringBuffer();
  while (!token.isEof) {
    a.write(' $token');
    token = token.next;
  }
}

Constructor scanTokens(Token token) {
  // StringBuffer stringBuffer = StringBuffer();
  StackImpl stackImpl = StackImpl();
  Token begin = token;
  token = token.next;
  // stringBuffer.write(' $token');
  if (token is BeginToken) {
    stackImpl.push(token.next);
  }
  // PropertyPosition uuidPosition;
  // PropertyPosition namePosition;
  List<PropertyPosition> uuidList = [];
  List<PropertyPosition> nameList = [];
  List<Property> properties = [];

  while (!token.isEof && stackImpl.isNotEmpty && stackImpl.arrayLength >= 1) {
    token = token.next;
    if (token is BeginToken && token.lexeme == '(') {
      ;
      stackImpl.push(token);
    }
    if (token.lexeme == ')') {
      stackImpl.pop();
    }
    if (isEndWithUuidToken(token)) {
      uuidList.add(PropertyPosition(token, token.next.next));
    }
    if (isEndWithNameToken(token)) {
      nameList.add(PropertyPosition(token, token.next.next));
    }
  }

  uuidList.forEach((element) {
    bool isMatch = false;
    nameList.forEach((element2) {
      if (isMatchUuidWithName(element2.beginToken, element.beginToken)) {
        isMatch = true;
        properties.add(Property()
          ..uuidPosition = element
          ..namePosition = element2);
      }
    });
    if (!isMatch) {
      if (element.beginToken.lexeme == 'uuid') {
        properties.add(Property()..uuidPosition = element);
      }
    }
  });
  return Constructor()
    ..properties = properties
    ..constructorBegin = begin;
}

bool isContainUUidToken(Token token) {
  return token.toString() == 'uuid' && token.next.type == TokenType.COLON && token.next.next.type == TokenType.STRING;
}

bool isContainNameToken(Token token) {
  return token.toString() == 'name' && token.next.type == TokenType.COLON && token.next.next.type == TokenType.STRING;
}

bool isEndWithUuidToken(Token token) {
  if (token.lexeme.length < 4) {
    return false;
  }
  String string = token.lexeme.substring(token.lexeme.length - 4, token.lexeme.length).toLowerCase();
  return string == 'uuid' && token.next.type == TokenType.COLON && token.next.next.type == TokenType.STRING;
}

bool isEndWithNameToken(Token token) {
  if (token.lexeme.length < 4) {
    return false;
  }
  String string = token.lexeme.substring(token.lexeme.length - 4, token.lexeme.length).toLowerCase();
  return string == 'name' && token.next.type == TokenType.COLON && token.next.next.type == TokenType.STRING;
}

bool isMatchUuidWithName(Token nameToken, Token UuidToken) {
  if (nameToken.lexeme.length < 4) {
    return false;
  }
  if (UuidToken.lexeme.length < 4) {
    return false;
  }
  String nameString = nameToken.lexeme.substring(0, nameToken.lexeme.length - 4).toLowerCase();
  String uuidString = UuidToken.lexeme.substring(0, UuidToken.lexeme.length - 4).toLowerCase();
  return nameString == uuidString;
}

Future<Map<String, dynamic>> reportJson(String path) async {
  if (path == null || path.isEmpty) {
    return null;
  }
  File file = File(path);
  if (!(await file.exists())) {
    return null;
  }
  Map<String, dynamic> map = Map<String, dynamic>.from(jsonDecode(await file.readAsString()));
  return map;
}
